import math, random

import gym
import gym.wrappers
import numpy as np

import sys
import torch
import torch.nn as nn
import torch.optim as optim
import torch.autograd as autograd 
import torch.nn.functional as F
from torch.distributions import Categorical

from collections import namedtuple

import sys
sys.path.append("..")
sys.path.append("../..")

import general_energy_nets.predictive_coding as pc

from IPython.display import clear_output
import matplotlib.pyplot as plt

from value_iteration import value_iteration, plot_grid

import argparse

ENV_ID = "FrozenLake-v1"
NUM_SEEDS = 10

PC_LR = 0.001
PC_OPTIM = 'Adam'

BP_LR = 0.001
BP_OPTIM = 'Adam'

BATCH_SIZE = 100
PERCENTILE = 30
GAMMA = 0.9

NESTEROV = False

MAX_ITERS = 100

Episode = namedtuple('Episode', field_names=['reward', 'steps'])
EpisodeStep = namedtuple('EpisodeStep', field_names=['observation', 'action'])

class NN(nn.Module):
    def __init__(self, num_inputs, num_actions, predictive_coding=False):
        super(NN, self).__init__()
        
        self.layers = []
        self.layers.append(nn.Linear(num_inputs, 128))
        if predictive_coding:
            self.layers.append(pc.PCLayer())
        self.layers.append(nn.ReLU())
        self.layers.append(nn.Linear(128, 128))
        if predictive_coding:
            self.layers.append(pc.PCLayer())
        self.layers.append(nn.ReLU())
        self.layers.append(nn.Linear(128, num_actions))
        self.layers = nn.Sequential(*self.layers)
        
    def forward(self, x):
        return self.layers(x)
    
    @torch.no_grad()
    def get_policy(self):
        policy = torch.zeros(env.nS, env.nA)

        for s in range(env.nS):
            _s = F.one_hot(torch.tensor(s), num_classes=env.observation_space.n).to(torch.float32).unsqueeze(0).to(device)
            logits = self.forward(_s)
            probs = F.softmax(logits, dim=1).cpu()
            policy[s] = probs
            
        return policy
    
    @torch.no_grad()
    def act(self, state):
        self.train(False)
        state = F.one_hot(torch.tensor(state), num_classes=env.observation_space.n).to(torch.float32).unsqueeze(0).to(device)
        logits = self.forward(state)
        probs = F.softmax(logits, dim=1)
        dist = Categorical(probs)
        action = dist.sample().cpu().numpy()[0]
        return action

def step_update(net, trainer, state, action):
    net.train()
    
    results = trainer.train_on_batch(
        inputs=state,
        loss_fn=cross_entropy_loss_fn,
        loss_fn_kwargs={
            '_targets': action
        },
        is_log_progress=False,
        is_return_results_every_t=False,
        is_checking_after_callback_after_t=False,
    )
    
    return results


def iterate_batches(env, net, batch_size):
    batch = []
    episode_reward = 0.0
    episode_steps = []
    obs = env.reset()
    while True:
        action = net.act(obs)
        next_obs, reward, is_done, _ = env.step(action)
        episode_reward += reward
        episode_steps.append(EpisodeStep(observation=obs, action=action))
        if is_done:
            batch.append(Episode(reward=episode_reward, steps=episode_steps))
            episode_reward = 0.0
            episode_steps = []
            next_obs = env.reset()
            if len(batch) == batch_size:
                yield batch
                batch = []
        obs = next_obs

def filter_batch(batch, percentile):
    disc_rewards = list(map(lambda s: s.reward * (GAMMA ** len(s.steps)), batch))
    reward_bound = np.percentile(disc_rewards, percentile)

    train_obs = []
    train_act = []
    elite_batch = []
    for example, discounted_reward in zip(batch, disc_rewards):
        if discounted_reward > reward_bound:
            train_obs.extend(map(lambda step: step.observation, example.steps))
            train_act.extend(map(lambda step: step.action, example.steps))
            elite_batch.append(example)

    return elite_batch, train_obs, train_act, reward_bound

def cross_entropy_loss_fn(output, _targets):
    return nn.CrossEntropyLoss()(output, _targets)

def compute_confidence_intervals(matrix):
    upper_confidence_interval = []
    lower_confidence_interval = []

    for iteration in matrix.transpose():
        bootstrapped_means = []
        for B in range(2000):
            bootstrapped_sample = np.random.choice(iteration, len(iteration))
            bootstrapped_means.append(np.mean(bootstrapped_sample))
        bootstrapped_percentiles = np.percentile(np.array(bootstrapped_means), [2.5, 97.5])
        upper_confidence_interval.append(bootstrapped_percentiles[1])
        lower_confidence_interval.append(bootstrapped_percentiles[0])

    return upper_confidence_interval, lower_confidence_interval

def plot(bp_policy_loss_matrix, bp_reward_matrix, pc_policy_loss_matrix, pc_reward_matrix, confidence_intervals=False):
    plt.style.use('ggplot')

    max_iters = max([len(policy_losses) for policy_losses in bp_policy_loss_matrix])

    bp_policy_loss_matrix = np.array([(policy_losses + [policy_losses[-1]]*(max_iters - len(policy_losses))) for policy_losses in bp_policy_loss_matrix])
    bp_reward_matrix = np.array([(value_losses + [value_losses[-1]]*(max_iters - len(value_losses)))  for value_losses in bp_reward_matrix])

    bp_policy_loss_mean = np.mean(bp_policy_loss_matrix, axis=0)
    bp_reward_mean = np.mean(bp_reward_matrix, axis=0)

    max_iters = max([len(policy_losses) for policy_losses in pc_policy_loss_matrix])

    pc_policy_loss_matrix = np.array([policy_losses + [policy_losses[-1]]*(max_iters - len(policy_losses)) for policy_losses in pc_policy_loss_matrix])
    pc_reward_matrix = np.array([value_losses + [value_losses[-1]]*(max_iters - len(value_losses)) for value_losses in pc_reward_matrix])

    pc_policy_loss_mean = np.mean(pc_policy_loss_matrix, axis=0)
    pc_reward_mean = np.mean(pc_reward_matrix, axis=0)

    if confidence_intervals:
        bp_policy_loss_upper_interval, bp_policy_loss_lower_interval = compute_confidence_intervals(bp_policy_loss_matrix)
        bp_reward_upper_interval, bp_reward_lower_interval = compute_confidence_intervals(bp_reward_matrix)

        pc_policy_loss_upper_interval, pc_policy_loss_lower_interval = compute_confidence_intervals(pc_policy_loss_matrix)
        pc_reward_upper_interval, pc_reward_lower_interval = compute_confidence_intervals(pc_reward_matrix)

    # plot each individually
    # plot BP on its own
    plt.figure(figsize=(20,5))
    plt.subplot(131)
    plt.title('BP Cross Entropy logistic policy losses')
    plt.plot([x for x in range(bp_policy_loss_matrix.shape[1])], bp_policy_loss_mean, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    if confidence_intervals:
        plt.fill_between([x for x in range(bp_policy_loss_matrix.shape[1])], bp_policy_loss_upper_interval, bp_policy_loss_lower_interval, color='blueviolet', alpha=0.4)

    plt.subplot(132)
    plt.title('BP Cross Entropy rewards')
    plt.plot([x for x in range(bp_reward_matrix.shape[1])], bp_reward_mean, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    if confidence_intervals:
        plt.fill_between([x for x in range(bp_reward_matrix.shape[1])], bp_reward_upper_interval, bp_reward_lower_interval, color='blueviolet', alpha=0.4)

    plt.savefig('./imgs/{}_bp_seeds={}_bp_lr={}_bp_optim={}_nesterov={}.png'.format(ENV_ID, NUM_SEEDS, BP_LR, BP_OPTIM, NESTEROV))
    plt.close('all')

    # plot PC on its own
    plt.figure(figsize=(20,5))
    plt.subplot(131)
    plt.title('PC Cross Entropy logistic policy losses')
    plt.plot([x for x in range(pc_policy_loss_matrix.shape[1])], pc_policy_loss_mean, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    if confidence_intervals:
        plt.fill_between([x for x in range(pc_policy_loss_matrix.shape[1])], pc_policy_loss_upper_interval, pc_policy_loss_lower_interval, color='firebrick', alpha=0.4)

    plt.subplot(132)
    plt.title('PC Cross Entropy rewards')
    plt.plot([x for x in range(pc_reward_matrix.shape[1])], pc_reward_mean, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    if confidence_intervals:
        plt.fill_between([x for x in range(pc_reward_matrix.shape[1])], pc_reward_upper_interval, pc_reward_lower_interval, color='firebrick', alpha=0.4)

    plt.savefig('./imgs/{}_pc_seeds={}_pc_lr={}_pc_optim={}_nesterov={}.png'.format(ENV_ID, NUM_SEEDS, PC_LR, PC_OPTIM, NESTEROV))
    plt.close('all')

    # plot both on same axis
    plt.figure(figsize=(20,5))
    plt.subplot(131)
    plt.title('logistic policy losses')
    plt.plot([x for x in range(bp_policy_loss_matrix.shape[1])], bp_policy_loss_mean, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    plt.plot([x for x in range(pc_policy_loss_matrix.shape[1])], pc_policy_loss_mean, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    plt.legend()
    
    if confidence_intervals:
        plt.fill_between([x for x in range(bp_policy_loss_matrix.shape[1])], bp_policy_loss_upper_interval, bp_policy_loss_lower_interval, color='blueviolet', alpha=0.4)
        plt.fill_between([x for x in range(pc_policy_loss_matrix.shape[1])], pc_policy_loss_upper_interval, pc_policy_loss_lower_interval, color='firebrick', alpha=0.4)

    plt.subplot(132)
    plt.title('rewards')
    plt.plot([x for x in range(bp_reward_matrix.shape[1])], bp_reward_mean, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    plt.plot([x for x in range(pc_reward_matrix.shape[1])], pc_reward_mean, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    plt.legend()
    
    if confidence_intervals:
        plt.fill_between([x for x in range(bp_reward_matrix.shape[1])], bp_reward_upper_interval, bp_reward_lower_interval, color='blueviolet', alpha=0.4)
        plt.fill_between([x for x in range(pc_reward_matrix.shape[1])], pc_reward_upper_interval, pc_reward_lower_interval, color='firebrick', alpha=0.4)

    plt.savefig('./imgs/{}_seeds={}_batch_size={}_percentile={}_bp_lr={}_bp_optim={}_pc_lr={}_pc_lr={}_nesterov={}.png'.format(ENV_ID, NUM_SEEDS, BATCH_SIZE, PERCENTILE, BP_LR, BP_OPTIM, PC_LR, PC_OPTIM, NESTEROV))
    plt.close('all')

def save_data(bp_policy_loss_matrix, pc_policy_loss_matrix, bp_reward_matrix, pc_reward_matrix):

    bp_mean_policy_loss = np.mean(np.mean(bp_policy_loss_matrix, axis=0))
    pc_mean_policy_loss = np.mean(np.mean(pc_policy_loss_matrix, axis=0))

    bp_mean_reward = np.mean(np.mean(bp_reward_matrix, axis=0))
    pc_mean_reward = np.mean(np.mean(pc_reward_matrix, axis=0))

    with open('./data/{}_seeds={}_batch_size={}_percentile={}_bp_lr={}_bp_optim={}_pc_lr={}_pc_lr={}_nesterov={}.txt'.format(ENV_ID, NUM_SEEDS, BATCH_SIZE, PERCENTILE, BP_LR, BP_OPTIM, PC_LR, PC_OPTIM, NESTEROV), 'w') as outfile:
        outfile.write('BP average solve time {:.2f}\n'.format(bp_mean_time_to_solve))
        outfile.write('PC average solve time {:.2f}\n'.format(pc_mean_time_to_solve))
        outfile.write('BP Mean Policy Loss ({}) : \n'.format(np.mean(bp_mean_policy_loss)))
        outfile.write(str(bp_mean_policy_loss))
        outfile.write('\n')
        outfile.write('PC Mean Policy Loss ({}) : \n'.format(np.mean(pc_mean_policy_loss)))
        outfile.write(str(pc_mean_policy_loss))
        outfile.write('\n')
        outfile.write('BP Mean Value Loss ({}) : \n'.format(np.mean(bp_mean_value_loss)))
        outfile.write(str(bp_mean_reward))
        outfile.write('\n')
        outfile.write('PC Mean Value Loss ({}) : \n'.format(np.mean(pc_mean_value_loss)))
        outfile.write(str(pc_mean_reward))
        outfile.write('\n')

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--n-seed", default=NUM_SEEDS)
    parser.add_argument("--env-id", default=ENV_ID)
    parser.add_argument("--batch-size", default=BATCH_SIZE)
    parser.add_argument("--percentile", default=PERCENTILE)
    parser.add_argument("--bp-lr", default=BP_LR)
    parser.add_argument("--pc-lr", default=PC_LR)
    parser.add_argument("--bp-optim", default=BP_OPTIM)
    parser.add_argument("--pc-optim", default=PC_OPTIM)
    parser.add_argument("--nesterov", default=NESTEROV)
    parser.add_argument("--max-iters", default=MAX_ITERS)
    args = parser.parse_args()

    NUM_SEEDS = int(args.n_seed)
    ENV_ID = args.env_id
    BATCH_SIZE = int(args.batch_size)
    PERCENTILE = int(args.percentile)
    BP_LR = float(args.bp_lr)
    PC_LR = float(args.pc_lr)
    BP_OPTIM = args.bp_optim
    PC_OPTIM = args.pc_optim
    NESTEROV = bool(int(args.nesterov))
    MAX_ITERS = int(args.max_iters)

    assert torch.cuda.is_available()
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu" )

    env = gym.make(ENV_ID, is_slippery=False) 
    env = gym.wrappers.TimeLimit(env, max_episode_steps=100)

    policy_optim, V_optim = value_iteration(env, gamma=GAMMA)
    policy_optim = torch.FloatTensor(policy_optim)

    # run BP
    bp_loss_matrix = []
    bp_reward_matrix = []

    for seed in range(742, 742+NUM_SEEDS):
        env = gym.make(ENV_ID, is_slippery=False) 
        env = gym.wrappers.TimeLimit(env, max_episode_steps=100)
        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        env.action_space.seed(seed)
        env.seed(seed)

        net = NN(env.observation_space.n, env.action_space.n, predictive_coding=False).to(device)
        net.eval()

        optimizer = None
        p_kwargs = None
        if BP_OPTIM == 'Adam':
            optimizer = optim.Adam
            p_kwargs = {"lr": BP_LR}
        elif BP_OPTIM == 'SGD':
            optimizer = optim.SGD
            if NESTEROV:
                p_kwargs = {"lr": BP_LR, "momentum" : 0.9, "nesterov" : True}
            else:
                p_kwargs = {"lr": BP_LR}
        else:
            raise ValueError("Bad argument for optim")

        bp_trainer = pc.PCTrainer(
            net,
            T=1,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": 0.1},
            x_lr_discount=0.5,
            update_p_at='last',
            optimizer_p_fn=optimizer,
            optimizer_p_kwargs=p_kwargs,
            plot_progress_at=[],
        )

        all_reward_means = []
        losses = []
        policy_losses = []

        full_batch = []
        for eps_idx, batch in enumerate(iterate_batches(env, net, BATCH_SIZE)):
            reward_mean = float(np.mean(list(map(lambda s: s.reward, batch))))

            full_batch, obs, acts, reward_bound = filter_batch(full_batch + batch, PERCENTILE)
            if not full_batch:
                continue

            obs_v = F.one_hot(torch.tensor(obs).to(torch.int64), num_classes=env.observation_space.n).to(torch.float32).to(device)
            acts_v = torch.LongTensor(acts).to(device)
            
            full_batch = full_batch[-500:]

            loss = step_update(net, bp_trainer, obs_v, acts_v)['loss'][0]
            losses.append(loss)
            
            net.eval()
            policy = net.get_policy()
            policy_loss = nn.CrossEntropyLoss()(policy, policy_optim).item()
            policy_losses.append(policy_loss)

            all_reward_means.append(reward_mean)
            
            if eps_idx > MAX_ITERS or reward_mean > 0.8:
                break

        print("Seed {} complete".format(seed))
        bp_loss_matrix.append(losses)
        bp_reward_matrix.append(all_reward_means)

    # run PC

    pc_loss_matrix = []
    pc_reward_matrix = []

    for seed in range(742, 742+NUM_SEEDS):
        env = gym.make(ENV_ID, is_slippery=False) 
        env = gym.wrappers.TimeLimit(env, max_episode_steps=100)
        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        env.action_space.seed(seed)
        env.seed(seed)

        net = NN(env.observation_space.n, env.action_space.n, predictive_coding=True).to(device)
        net.eval()

        optimizer = None
        p_kwargs = None
        if PC_OPTIM == 'Adam':
            optimizer = optim.Adam
            p_kwargs = {"lr": PC_LR}
        elif PC_OPTIM == 'SGD':
            optimizer = optim.SGD
            if NESTEROV:
                p_kwargs = {"lr": PC_LR, "momentum" : 0.9, "nesterov" : True}
            else:
                p_kwargs = {"lr": PC_LR}
        else:
            raise ValueError("Bad argument for optim")

        pc_trainer = pc.PCTrainer(
            net,
            T=128,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": 0.1},
            x_lr_discount=0.5,
            update_p_at='last',
            optimizer_p_fn=optimizer,
            optimizer_p_kwargs=p_kwargs,
            plot_progress_at=[],
        )

        all_reward_means = []
        losses = []
        policy_losses = []

        full_batch = []
        for eps_idx, batch in enumerate(iterate_batches(env, net, BATCH_SIZE)):
            reward_mean = float(np.mean(list(map(lambda s: s.reward, batch)))) 

            full_batch, obs, acts, reward_bound = filter_batch(full_batch + batch, PERCENTILE)
            if not full_batch:
                continue

            obs_v = F.one_hot(torch.tensor(obs).to(torch.int64), num_classes=env.observation_space.n).to(torch.float32).to(device)
            acts_v = torch.LongTensor(acts).to(device)
            
            full_batch = full_batch[-500:]

            loss = step_update(net, pc_trainer, obs_v, acts_v)['loss'][0]
            losses.append(loss)
            
            net.eval()
            policy = net.get_policy()
            policy_loss = nn.CrossEntropyLoss()(policy, policy_optim).item()
            policy_losses.append(policy_loss)

            all_reward_means.append(reward_mean)
            
            if eps_idx > MAX_ITERS or reward_mean > 0.8:
                break

        print("Seed {} complete".format(seed))
        pc_loss_matrix.append(losses)
        pc_reward_matrix.append(all_reward_means)

    plot(bp_loss_matrix, bp_reward_matrix, pc_loss_matrix, pc_reward_matrix, confidence_intervals=(NUM_SEEDS >= 3))